import 'package:flutter/widgets.dart';
import 'element_locator_utils.dart';

const _kTargetableWidgets = [
  'TextButton',
  'ElevatedButton',
  'FloatingActionButton',
  'IconButton',
  'InkWell',
  'GestureDetector',
  'DropdownButton',
  'OutlinedButton',
  'Radio',
  'PopupMenuButton',
];

/// ElementLocator is responsible for retrieving from the element tree all leaf elements which location matches the given position
class ElementLocator {
  late Offset _currentTouchPosition;
  final Map<String, int> _elementPathOccurrences = {};
  final List<String> _targetedElementsPaths = [];

  /// Construct the ElementLocator with the given element & position and search the element tree for leaf elements matching the given position
  ElementLocator();

  /// Scan the element tree for candidate elements at the given position
  String scanForCandidateElementsAtPosition(Offset position) {
    final rootElement = WidgetsBinding.instance!.renderViewElement!;
    _currentTouchPosition = position;
    _visitSubtree(rootElement, 'Root');
    return _getTargetElementPath();
  }

  static bool _isWidgetTargetable(Element element) {
    return _kTargetableWidgets.contains(element.widget.runtimeType.toString());
  }

  static bool _isWithinRect({required Offset position, required Rect rect}) =>
      (rect.left < position.dx &&
          position.dx < rect.right &&
          rect.top < position.dy &&
          position.dy < rect.bottom);

  static Rect? _getElementLocation(Element element) {
    final renderObject = element.renderObject;
    final translation = renderObject?.getTransformTo(null).getTranslation();
    if (translation != null && renderObject?.paintBounds != null) {
      final globalPosition = (renderObject!.paintBounds
          .shift(Offset(translation.x, translation.y)));

      return globalPosition;
    }
  }

  void _updateElementPathOccurrences(String newPath) {
    if (_elementPathOccurrences.containsKey(newPath)) {
      _elementPathOccurrences[newPath] = _elementPathOccurrences[newPath]! + 1;
    } else {
      _elementPathOccurrences[newPath] = 0;
    }
  }

  /// We append element with their positional index to be able to distinguish elements with the same path (sibling elements)
  /// By counting the occurences of each element path (starting at 0) we can attribute a positional index based on the number of occurences (minus 1)
  /// If an element is a single child (its element path will be unique) -> eq(0)
  /// If an element has 3 siblings (they will all have the same path), so we respectively attribute the positional indexes: eq(0), eq(1), eq(2)
  String _appendPositionalIndexToPath(String path) {
    if (_elementPathOccurrences[path] == null) {
      /// _elementPathOccurrences[path] is updated by _updateElementPathOccurrences and should not end up null by the time it's called
      /// if for some reason it is though, we simply want to ignore this step and return the path unchanged
      return path;
    }
    return '$path:eq(${_elementPathOccurrences[path].toString()})';
  }

  void _visitSubtree(Element element, String path) {
    element.visitChildren((element) {
      if (isElementPrivate(element)) {
        /// if the element is private, we simply continue down the tree
        return _visitSubtree(element, path);
      }

      var newPath =
          '$path>${trimGenericParams(element.widget.runtimeType.toString())}';
      _updateElementPathOccurrences(newPath);
      newPath = _appendPositionalIndexToPath(newPath);

      if (element.toDiagnosticsNode().getChildren().isEmpty ||
          _isWidgetTargetable(element)) {
        final elementLocation = _getElementLocation(element);
        if (elementLocation == null) {
          return;
        }

        if (_isWithinRect(
            position: _currentTouchPosition, rect: elementLocation)) {
          _targetedElementsPaths.add(newPath);
          // we found a candidate element among the [_kTargetableWidgets]
          // located at the right position
          // we don't need to go deeper
          return;
        }
      }

      return _visitSubtree(element, newPath);
    });
  }

  /// Return the path of the element considered as the target of the touch
  String _getTargetElementPath() {
    return _targetedElementsPaths.last;
  }
}
