import 'package:contentsquare_platform_interface/contentsquare_platform_interface.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/widgets.dart';

const _kDragGestureDetectionMinDistance = 48.0;
const _kLongPressGestureDetectionMaxDistance = 24.0;
const _kFlickGestureMinVelocity = 100.0;

/// GestureListener
/// Detects all supported gesture and locate their position on the screen
class GestureListener extends StatefulWidget {
  /// Child widget
  final Widget child;

  /// Callback for the Tap gesture
  final Function(Offset position)? onTap;

  /// Callback for the onLongPress gesture
  final Function(Offset position)? onLongPress;

  /// Callback for the Tap gesture
  final Function(Offset position, CSGestureData data)? onDrag;

  /// Callback for the Tap gesture
  final Function(Offset position, CSGestureData data)? onFlick;

  /// GestureListener widget constructor
  const GestureListener({
    Key? key,
    required this.child,
    this.onTap,
    this.onLongPress,
    this.onDrag,
    this.onFlick,
  }) : super(key: key);

  @override
  GestureListenerState createState() => GestureListenerState();
}

/// GestureListenerState
class GestureListenerState extends State<GestureListener> {
  Offset? _horizontalDragGestureInitialPosition;
  Offset? _horizontalDragGestureCurrentPosition;
  Offset? _verticalDragGestureInitialPosition;
  Offset? _verticalDragGestureCurrentPosition;
  Offset? _longPressGestureInitialPosition;
  Offset? _longPressGestureCurrentPosition;

  bool _isFlickGesture(double primaryVelocity) {
    return primaryVelocity.abs() > _kFlickGestureMinVelocity;
  }

  CSGestureDirection _getHorizontalDragGestureDirection() {
    assert(_verticalDragGestureInitialPosition != null);
    assert(_verticalDragGestureCurrentPosition != null);

    return _getHorizontalDirection(_verticalDragGestureInitialPosition!,
        _verticalDragGestureCurrentPosition!);
  }

  CSGestureDirection _getVerticalDragGestureDirection() {
    assert(_verticalDragGestureInitialPosition != null);
    assert(_verticalDragGestureCurrentPosition != null);

    return _getVerticalDirection(_verticalDragGestureInitialPosition!,
        _verticalDragGestureCurrentPosition!);
  }

  void _handleHorizontalDragGestureStart(DragStartDetails details) {
    _horizontalDragGestureInitialPosition = details.globalPosition;
  }

  void _handleLongPressGestureStart(LongPressStartDetails details) {
    _longPressGestureInitialPosition = details.globalPosition;
  }

  void _handleVerticalDragGestureStart(DragStartDetails details) {
    _verticalDragGestureInitialPosition = details.globalPosition;
  }

  void _handleHorizontalDragGestureUpdate(DragUpdateDetails details) {
    _horizontalDragGestureCurrentPosition = details.globalPosition;
  }

  void _handleVerticalDragGestureUpdate(DragUpdateDetails details) {
    _verticalDragGestureCurrentPosition = details.globalPosition;
  }

  void _resetLongPressGestureTracking() {
    _longPressGestureInitialPosition = null;
    _longPressGestureCurrentPosition = null;
  }

  void _resetHorizontalDragGestureTracking() {
    _horizontalDragGestureInitialPosition = null;
    _horizontalDragGestureCurrentPosition = null;
  }

  void _resetVerticalDragGestureTracking() {
    _verticalDragGestureInitialPosition = null;
    _verticalDragGestureCurrentPosition = null;
  }

  double _getLongPressGestureDistance() {
    if (_longPressGestureInitialPosition == null ||
        _longPressGestureCurrentPosition == null) {
      return 0.0;
    }

    return _computeDistance(
      _longPressGestureCurrentPosition!,
      _longPressGestureInitialPosition!,
    );
  }

  double _getHorizontalDragGestureDistance() {
    if (_horizontalDragGestureInitialPosition == null ||
        _horizontalDragGestureCurrentPosition == null) {
      return 0.0;
    }

    return _computeHorizontalDistance(
      _horizontalDragGestureCurrentPosition!,
      _horizontalDragGestureInitialPosition!,
    );
  }

  double _getVerticalDragGestureDistance() {
    if (_verticalDragGestureInitialPosition == null ||
        _verticalDragGestureCurrentPosition == null) {
      return 0.0;
    }

    return _computeVerticalDistance(
      _verticalDragGestureCurrentPosition!,
      _verticalDragGestureInitialPosition!,
    );
  }

  bool _hasDetectedLongPressGesture() {
    return _getLongPressGestureDistance() <
        _kLongPressGestureDetectionMaxDistance;
  }

  bool _hasDetectedHorizontalDragGesture() {
    return _getHorizontalDragGestureDistance() >
        _kDragGestureDetectionMinDistance;
  }

  bool _hasDetectedVerticalDragGesture() {
    return _getVerticalDragGestureDistance() >
        _kDragGestureDetectionMinDistance;
  }

  static double _computeVerticalDistance(Offset point1, Offset point2) {
    return (point1 - point2).dy.abs();
  }

  static double _computeHorizontalDistance(Offset point1, Offset point2) {
    return (point1 - point2).dx.abs();
  }

  static double _computeDistance(Offset point1, Offset point2) {
    return (point1 - point2).distance;
  }

  static CSGestureDirection _getVerticalDirection(
      Offset point1, Offset point2) {
    return (point2 - point1).dy > 0.0
        ? CSGestureDirection.down
        : CSGestureDirection.up;
  }

  static CSGestureDirection _getHorizontalDirection(
      Offset point1, Offset point2) {
    return (point2 - point1).dx > 0.0
        ? CSGestureDirection.right
        : CSGestureDirection.left;
  }

  @override
  Widget build(BuildContext context) {
    return RawGestureDetector(
      gestures: {
        AllowMultipleTapGestureRecognizer: GestureRecognizerFactoryWithHandlers<
            AllowMultipleTapGestureRecognizer>(
          () => AllowMultipleTapGestureRecognizer(),
          (AllowMultipleTapGestureRecognizer instance) {
            instance.onTapUp = (TapUpDetails details) {
              if (widget.onTap != null) {
                widget.onTap!(details.globalPosition);
              }
            };
          },
        ),
        AllowMultipleLongPressGestureRecognizer:
            GestureRecognizerFactoryWithHandlers<
                AllowMultipleLongPressGestureRecognizer>(
          () => AllowMultipleLongPressGestureRecognizer(),
          (AllowMultipleLongPressGestureRecognizer instance) {
            instance.onLongPressStart = _handleLongPressGestureStart;
            instance.onLongPressEnd = (LongPressEndDetails details) {
              _longPressGestureCurrentPosition = details.globalPosition;
              if (_hasDetectedLongPressGesture()) {
                if (widget.onLongPress != null) {
                  widget.onLongPress!(details.globalPosition);
                }
                _resetLongPressGestureTracking();
              }
            };
          },
        ),
        AllowMultipleHorizontalDragGestureRecognizer:
            GestureRecognizerFactoryWithHandlers<
                AllowMultipleHorizontalDragGestureRecognizer>(
          () => AllowMultipleHorizontalDragGestureRecognizer(),
          (AllowMultipleHorizontalDragGestureRecognizer instance) {
            instance.dragStartBehavior = DragStartBehavior.down;
            instance.onStart = _handleHorizontalDragGestureStart;
            instance.onUpdate = _handleHorizontalDragGestureUpdate;
            instance.onEnd = (DragEndDetails details) {
              if (_hasDetectedHorizontalDragGesture() &&
                  details.primaryVelocity != null) {
                final swipeEventData = CSGestureData(
                  direction: _getHorizontalDragGestureDirection(),
                  distance: _getHorizontalDragGestureDistance().toInt(),
                  velocity: details.primaryVelocity!.toInt(),
                );

                if (_isFlickGesture(details.primaryVelocity!)) {
                  if (widget.onFlick != null) {
                    widget.onFlick!(
                        _horizontalDragGestureInitialPosition!, swipeEventData);
                  }
                } else {
                  if (widget.onDrag != null) {
                    widget.onDrag!(
                      _horizontalDragGestureInitialPosition!,
                      swipeEventData,
                    );
                  }
                }
                _resetHorizontalDragGestureTracking();
              }
            };
          },
        ),
        AllowMultipleVerticalDragGestureRecognizer:
            GestureRecognizerFactoryWithHandlers<
                AllowMultipleVerticalDragGestureRecognizer>(
          () => AllowMultipleVerticalDragGestureRecognizer(),
          (AllowMultipleVerticalDragGestureRecognizer instance) {
            instance.dragStartBehavior = DragStartBehavior.down;
            instance.onStart = _handleVerticalDragGestureStart;
            instance.onUpdate = _handleVerticalDragGestureUpdate;
            instance.onEnd = (DragEndDetails details) {
              if (_hasDetectedVerticalDragGesture() &&
                  details.primaryVelocity != null) {
                final swipeEventData = CSGestureData(
                  direction: _getVerticalDragGestureDirection(),
                  distance: _getVerticalDragGestureDistance().toInt(),
                  velocity: details.primaryVelocity!.toInt(),
                );

                if (_isFlickGesture(details.primaryVelocity!)) {
                  if (widget.onFlick != null) {
                    widget.onFlick!(
                      _verticalDragGestureInitialPosition!,
                      swipeEventData,
                    );
                  }
                } else {
                  if (widget.onDrag != null) {
                    widget.onDrag!(
                      _verticalDragGestureInitialPosition!,
                      swipeEventData,
                    );
                  }
                }
                _resetVerticalDragGestureTracking();
              }
            };
          },
        ),
      },
      behavior: HitTestBehavior.translucent,
      child: widget.child,
    );
  }
}

/// AllowingMultipleGestureRecognizer mixin
/// rejectGesture() is overridden. When a gesture is rejected, this is the function that is called. By default, it disposes of the
/// Recognizer and runs clean up. However we modified it so that instead the Recognizer is disposed of, it is actually manually added.
/// The result is instead you have one Recognizer winning the Arena, you have two. It is a win-win.
/// This mixin can be used on any class inheriting from GestureRecognizer.
/// This way we can easily provide the capability of handling multiple recognizer on any kind of Gesture Recognizer
mixin AllowingMultipleGestureRecognizer on GestureRecognizer {
  @override
  void rejectGesture(int pointer) {
    acceptGesture(pointer);
    if (!runtimeType.toString().contains('Drag')) {
      super.rejectGesture(pointer);
    }
  }
}

/// AllowMultipleTapGestureRecognizer
/// Uses the AllowingMultipleGestureRecognizer mixin to provide the multiple gesture recognition capability
class AllowMultipleTapGestureRecognizer extends TapGestureRecognizer
    with AllowingMultipleGestureRecognizer {}

/// AllowMultipleLongPressGestureRecognizer
/// Uses the AllowingMultipleGestureRecognizer mixin to provide the multiple gesture recognition capability
class AllowMultipleLongPressGestureRecognizer extends LongPressGestureRecognizer
    with AllowingMultipleGestureRecognizer {}

/// AllowMultipleHorizontalDragGestureRecognizer
/// Uses the AllowingMultipleGestureRecognizer mixin to provide the multiple gesture recognition capability
class AllowMultipleHorizontalDragGestureRecognizer
    extends HorizontalDragGestureRecognizer
    with AllowingMultipleGestureRecognizer {}

/// AllowMultipleVerticalDragGestureRecognizer
/// Uses the AllowingMultipleGestureRecognizer mixin to provide the multiple gesture recognition capability
class AllowMultipleVerticalDragGestureRecognizer
    extends VerticalDragGestureRecognizer
    with AllowingMultipleGestureRecognizer {}
